#!/usr/bin/python3

import functools
from copy import deepcopy

from .grammar import BASE_NODE_TYPES


class NodeBase:
    """Represents a node within the solidity AST.

    Attributes:
        depth: Number of nodes between this node and the SourceUnit
        offset: Absolute source offsets as a (start, stop) tuple
        contract_id: Contract ID as given by the standard compiler JSON
        fields: List of attributes for this node
    """

    def __init__(self, ast, parent):
        self.depth = parent.depth + 1 if parent is not None else 0
        self._parent = parent
        self._children = set()
        src = [int(i) for i in ast["src"].split(":")]
        self.offset = (src[0], src[0] + src[1])
        self.contract_id = src[2]
        self.fields = sorted(ast.keys())

        for key, value in ast.items():
            if isinstance(value, dict) and value.get("nodeType") == "Block":
                value = value["statements"]
            elif key == "body" and not value:
                value = []
            if isinstance(value, dict):
                item = node_class_factory(value, self)
                if isinstance(item, NodeBase):
                    self._children.add(item)
                setattr(self, key, item)
            elif isinstance(value, list):
                items = [node_class_factory(i, self) for i in value]
                setattr(self, key, items)
                self._children.update(i for i in items if isinstance(i, NodeBase))
            else:
                setattr(self, key, value)

    def __hash__(self):
        return hash(f"{self.nodeType}{self.depth}{self.offset}")

    def __repr__(self):
        repr_str = f"<{self.nodeType}"
        if hasattr(self, "nodes"):
            repr_str += " iterable"
        if hasattr(self, "type"):
            if isinstance(self.type, str):
                repr_str += f" {self.type}"
            else:
                repr_str += f" {self.type._display()}"
        if self._display():
            repr_str += f" '{self._display()}'"
        else:
            repr_str += " object"
        return f"{repr_str}>"

    def _display(self):
        if hasattr(self, "name") and hasattr(self, "value"):
            return f"{self.name} = {self.value}"
        for attr in ("name", "value", "absolutePath"):
            if hasattr(self, attr):
                return f"{getattr(self, attr)}"
        return ""

    def children(
        self,
        depth=None,
        include_self=False,
        include_parents=True,
        include_children=True,
        required_offset=None,
        offset_limits=None,
        filters=None,
        exclude_filter=None,
    ):
        """Get childen nodes of this node.

        Arguments:
          depth: Number of levels of children to traverse. 0 returns only this node.
          include_self: Includes this node in the results.
          include_parents: Includes nodes that match in the results, when they also have
                        child nodes that match.
          include_children: If True, as soon as a match is found it's children will not
                            be included in the search.
          required_offset: Only match nodes with a source offset that contains this offset.
          offset_limits: Only match nodes when their source offset is contained inside
                           this source offset.
          filters: Dictionary of {attribute: value} that children must match. Can also
                   be given as a list of dicts, children that match one of the dicts
                   will be returned.
          exclude_filter: Dictionary of {attribute:value} that children cannot match.

        Returns:
            List of node objects."""
        if filters is None:
            filters = {}
        if exclude_filter is None:
            exclude_filter = {}
        if isinstance(filters, dict):
            filters = [filters]
        filter_fn = functools.partial(
            _check_filters, required_offset, offset_limits, filters, exclude_filter
        )
        find_fn = functools.partial(_find_children, filter_fn, include_parents, include_children)
        result = find_fn(find_fn, depth, self)
        if include_self or not result or result[0] != self:
            return result
        return result[1:]

    def parents(self, depth=-1, filters=None):
        """Get parent nodes of this node.

        Arguments:
            depth: Depth limit. If given as a negative value, it will be subtracted
                   from this object's depth.
            filters: Dictionary of {attribute: value} that parents must match.

        Returns: list of nodes"""
        if filters and not isinstance(filters, dict):
            raise TypeError("Filters must be a dict")
        if depth < 0:
            depth = self.depth + depth
        if depth >= self.depth or depth < 0:
            raise IndexError("Given depth exceeds node depth")
        node_list = []
        parent = self
        while True:
            parent = parent._parent
            if not filters or _check_filter(parent, filters, {}):
                node_list.append(parent)
            if parent.depth == depth:
                return node_list

    def parent(self, depth=-1, filters=None):
        """Get a parent node of this node.

        Arguments:
            depth: Depth limit. If given as a negative value, it will be subtracted
                   from this object's depth. The parent at this exact depth is returned.
            filters: Dictionary of {attribute: value} that the parent must match.

        If a filter value is given, will return the first parent that meets the filters
        up to the given depth. If none is found, returns None.

        If no filter is given, returns the parent at the given depth."""
        if filters and not isinstance(filters, dict):
            raise TypeError("Filters must be a dict")
        if depth < 0:
            depth = self.depth + depth
        if depth >= self.depth or depth < 0:
            raise IndexError("Given depth exceeds node depth")
        parent = self
        while parent.depth > depth:
            parent = parent._parent
            if parent.depth == depth and not filters:
                return parent
            if filters and _check_filter(parent, filters, {}):
                return parent
        return None

    def is_child_of(self, node):
        """Checks if this object is a child of the given node object."""
        if node.depth >= self.depth:
            return False
        return self.parent(node.depth) == node

    def is_parent_of(self, node):
        """Checks if this object is a parent of the given node object."""
        if node.depth <= self.depth:
            return False
        return node.parent(self.depth) == self

    def get(self, key, default=None):
        """
        Gets an attribute from this node, if that attribute exists.

        Arguments:
            key: Field name to return. May contain decimals to return a value
                 from a child node.
            default: Default value to return.

        Returns: Field value if it exists. Default value if not.
        """
        if key is None:
            raise TypeError("Cannot match against None")
        obj = self
        for k in key.split("."):
            if isinstance(obj, dict):
                obj = obj.get(k)
            else:
                obj = getattr(obj, k, None)
        return obj or default


class IterableNodeBase(NodeBase):
    def __getitem__(self, key):
        if isinstance(key, str):
            try:
                return next(i for i in self.nodes if getattr(i, "name", None) == key)
            except StopIteration:
                raise KeyError(key)
        return self.nodes[key]

    def __iter__(self):
        return iter(self.nodes)

    def __len__(self):
        return len(self.nodes)

    def __contains__(self, obj):
        return obj in self.nodes


def node_class_factory(ast, parent):
    ast = deepcopy(ast)
    if not isinstance(ast, dict) or "nodeType" not in ast:
        return ast
    if "body" in ast:
        ast["nodes"] = ast.pop("body")
    base_class = IterableNodeBase if "nodes" in ast else NodeBase
    base_type = next((k for k, v in BASE_NODE_TYPES.items() if ast["nodeType"] in v), None)
    if base_type:
        ast["baseNodeType"] = base_type
    return type(ast["nodeType"], (base_class,), {})(ast, parent)


def _check_filters(required_offset, offset_limits, filters, exclude, node):
    if required_offset and not is_inside_offset(required_offset, node.offset):
        return False
    if offset_limits and not is_inside_offset(node.offset, offset_limits):
        return False
    for f in filters:
        if _check_filter(node, f, exclude):
            return True
    return False


def _check_filter(node, filters, exclude):
    for key, value in filters.items():
        if node.get(key) != value:
            return False
    for key, value in exclude.items():
        if node.get(key) == value:
            return False
    return True


def _find_children(filter_fn, include_parents, include_children, find_fn, depth, node):
    if depth is not None:
        depth -= 1
        if depth < 0:
            return [node] if filter_fn(node) else []
    if not include_children and filter_fn(node):
        return [node]
    node_list = []
    for child in node._children:
        node_list.extend(find_fn(find_fn, depth, child))
    if (include_parents or not node_list) and filter_fn(node):
        node_list.insert(0, node)
    return node_list


def is_inside_offset(inner, outer):
    """Checks if the first offset is contained in the second offset

    Args:
        inner: inner offset tuple
        outer: outer offset tuple

    Returns: bool"""
    return outer[0] <= inner[0] <= inner[1] <= outer[1]



#!/usr/bin/python3


def set_dependencies(source_nodes):
    """Sets contract node dependencies.

    Arguments:
        source_nodes: list of SourceUnit objects.

    Returns: SourceUnit objects where all ContractDefinition nodes contain
             'dependencies' and 'libraries' attributes."""
    symbol_map = get_symbol_map(source_nodes)
    contract_list = [x for i in source_nodes for x in i if x.nodeType == "ContractDefinition"]

    # add immediate dependencies
    for contract in contract_list:
        contract.dependencies = set()
        contract.libraries = dict()

        # listed dependencies
        for key in contract.contractDependencies:
            contract.dependencies.add(symbol_map[key])

        # using .. for libraries
        for node in contract.children(filters={"nodeType": "UsingForDirective"}):
            ref_node = symbol_map[node.libraryName.referencedDeclaration]
            contract.libraries[node.libraryName.name] = ref_node
            contract.dependencies.add(ref_node)

        # imported contracts used as types in assignment
        for node in contract.children(filters={"nodeType": "UserDefinedTypeName"}):
            ref_id = node.referencedDeclaration
            if ref_id in symbol_map:
                contract.dependencies.add(symbol_map[ref_id])

        # imported contracts as types, no assignment
        for node in contract.children(
            filters={"nodeType": "FunctionCall", "expression.nodeType": "Identifier"}
        ):
            if node.typeDescriptions["typeString"].startswith("contract "):
                ref_id = node.expression.referencedDeclaration
                if ref_id in symbol_map:
                    contract.dependencies.add(symbol_map[ref_id])

        # unlinked libraries
        for node in contract.children(filters={"nodeType": "Identifier"}):
            ref_node = symbol_map.get(node.referencedDeclaration)
            if ref_node is None:
                continue
            if ref_node.nodeType in ("EnumDefinition", "StructDefinition"):
                contract.dependencies.add(ref_node)
            if ref_node.nodeType == "ContractDefinition" and ref_node.contractKind == "library":
                contract.dependencies.add(ref_node)

        # prevent recursion errors from self-dependency
        contract.dependencies.discard(contract)

    # add dependencies of dependencies
    for contract in contract_list:
        current_deps = contract.dependencies

        while True:
            expanded_deps = set(x for i in current_deps for x in getattr(i, "dependencies", []))
            expanded_deps |= current_deps
            expanded_deps.discard(contract)

            if current_deps == expanded_deps:
                break
            current_deps = expanded_deps

        current_deps |= {symbol_map[i] for i in contract.linearizedBaseContracts}

        contract.dependencies = current_deps

    # convert dependency sets to lists
    for contract in contract_list:
        if contract in contract.dependencies:
            # a contract should not list itself as a dependency
            contract.dependencies.remove(contract)

        contract.dependencies = sorted(contract.dependencies, key=lambda k: k.name)

    return source_nodes


def get_symbol_map(source_nodes):
    """Generates a dict of {'id': SourceUnit} used for linking nodes.

    Arguments:
        source_nodes: list of SourceUnit objects."""
    symbol_map = {}
    for node in source_nodes:
        for key, value in ((k, x) for k, v in node.exportedSymbols.items() for x in v):
            try:
                symbol_map[value] = node[key]
            except KeyError:
                # solc >=0.7.2 may include exported symbols that reference
                # other contracts, handle this gracefully
                pass

    return symbol_map



#!/usr/bin/python3

import json
from copy import deepcopy
from pathlib import Path

import pytest

import solcast

JSON_PATHS = list(Path("tests/compiled").glob("*.json"))


def test_dependencies():
    ast = solcast.from_standard_output_json("tests/compiled/aragon.json")
    node = next(i for i in ast[-1] if i.id == 16359)
    deps = set(i.name for i in node.dependencies)
    expected = {
        "Assert",
        "DelegateProxy",
        "ERCProxy",
        "IsContract",
        "ScriptHelpers",
        "Target",
        "ThrowProxy",
    }
    assert deps == expected


@pytest.mark.parametrize("path", JSON_PATHS)
def test_solcast(path):
    solcast.from_standard_output_json(path)


@pytest.mark.parametrize("path", JSON_PATHS)
def test_mutation(path):
    with open(path) as fp:
        output_json = json.load(fp)
    original = deepcopy(output_json)
    solcast.from_standard_output(output_json)
    assert original == output_json


def test_dependencies_sol0_8_6():
    result = solcast.from_standard_output_json("tests/compiled/lp-token.json")
    lp_token_src = next(v for v in result if v.absolutePath == "./contracts/LpToken.sol")
    lp_token = lp_token_src["LpToken"]
    assert any(dep.name == "ERC20" for dep in lp_token.dependencies)



#!/usr/bin/env python

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="py-solc-ast",
    version="1.2.10",
    description="A tool for exploring the abstract syntax tree generated by solc.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Ben Hauser",
    author_email="ben@hauser.id",
    url="https://github.com/iamdefinitelyahuman/py-solc-ast",
    include_package_data=True,
    py_modules=["solcast"],
    setup_requires=[],
    python_requires=">=3.6, <4",
    install_requires=[],
    license="MIT",
    zip_safe=False,
    keywords=["ethereum", "solidity", "solc", "ast"],
    packages=find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)
